#!/usr/bin/perl
#
#    nmis_logwatch.pl - NMIS Perl Program - Network Mangement Information System
#    Copyright (C) 2000 Sinclair InterNetworking Services Pty Ltd
#    <nmis@sins.com.au> http://www.sins.com.au/nmis
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#*****************************************************************************
#
# written 20 sep 2002 eric.greenwood@imlmnetwork.com
# updated 16Aug2004
#
#****** Shouldn't be anything else to customise below here *******************
# best to customise in the nmis.conf file.
#
# run this from the <nmis>/bin directory
# turn on debug if you want detailed logging to the <nmis>/logs/nmis.log file
# ./logwatch.pl [debug=true]
# will fork and leave a status message so you know it is running.
#
# Auto configure to the <nmis-base>/lib and <nmis-base>/files/nmis.conf
use FindBin;
use lib "$FindBin::Bin/../lib";
use csv;
use strict;
use func;
use NMIS;
use POSIX;
use File::Tail;

# this imports the LOCK_ *constants (eg. LOCK_UN, LOCK_EX)
use Fcntl qw(:DEFAULT :flock);

my $pid;
my $line;

# Variables for command line munging
my %nvp = getArguements(@ARGV);

# See if customised config file required.
my $conf;
if ( $nvp{file} ne "" ) { $conf = $nvp{file}; }
else { $conf = "nmis.conf"; }
my $configfile = "$FindBin::Bin/../conf/$conf";
if ( -r $configfile ) { loadConfiguration($configfile); }
# the following should be conformant to Linux FHS
elsif ( -e "/etc/nmis/$conf") { loadConfiguration ("/etc/nmis/$conf");}
else { die "Can't access neither configuration file $configfile, nor /etc/nmis/$conf \n"; }

# Set debugging level.
my $debug = setDebug($nvp{debug});
my $vardir = $nvp{vardir};

unless ( $vardir ) { $vardir = $NMIS::config{'<nmis_var>'} }

# Initialize globals

my $restartcmd = "$0 '$vardir'";

# See if another instance of ourself is running, if so kill the process
my $pidfile = $vardir."/logwatch.pid";

if (-f $pidfile) {
  open(F, "<$pidfile");
  $pid = <F>;
  close(F);
  chomp $pid;
  if ($pid != $$) {
    logMessage("NMIS_LogWatch pidfile exists killing the pidfile process $pid");
    kill 9, $pid;
    unlink($pidfile);
    logMessage("NMIS_LogWatch pidfile $pidfile deleted");
  }
}

# As we are paranoid and running as root, chroot to a safe directory:
# note: if we go anywhere else that will break the file paths.
chroot("/") or die "Couldn't chroot to /: $!";


FORK: {
    if ($pid = fork) {
        # parent here
        # child process pid is available in $pid
		exit;	
    }
    elsif (defined $pid) { # $pid is zero here if defined
        # child here
        # parent process pid is available with getppid

		# dissociate from the controlling terminal
		POSIX::setsid() or die "Can't start new session: $!"; 
		# set our named so that rc scripts can figure out who we are.
		$0="NMIS_Logwatch debug=$debug";
		
		# Announce our presence via a PID file

		open(PID, ">$pidfile") || exit;
		print PID $$; close(PID);
		logMessage("NMIS_LogWatch pidfile $pidfile created");

		# Perform a sanity check. If the current PID file is not the same as
		# our PID then we have become detached somehow, so just exit

		open(PID, "<$pidfile") || exit;
		$pid = <PID>; close(PID);
		chomp $pid;
		exit unless $pid == $$;

		# Record our (re)starting in the event log

		logMessage("NMIS_LogWatch start: pidfile=$pidfile pid=$pid");

		# code body here.

		# setup a log file and trap any error messages as well.
		umask 0;
		open STDOUT, ">>$NMIS::config{log_root}/nmis.log" or die "Can't write to $NMIS::config{log_root}/nmis.log: $!";
		open STDERR, ">>$NMIS::config{log_root}/nmis.log" or die "Can't write to $NMIS::config{log_root}/nmis.log: $!";

		# setup a trap for fatal signals, setting a flag to indicate we need to gracefully exit.
		my $time_to_die = 0;
		$SIG{INT} = $SIG{TERM} = $SIG{HUP} = \&signal_handler;
		$SIG{'CHLD'} = 'IGNORE';		# autoreap zombies

		# check the intended tail file exists.
		my $Filename = "$NMIS::config{syslog_log}";
		my $File = File::Tail->new(name=>$Filename, tail=>0, maxinterval=>60, interval=>1.0);
		if (not defined $File) {
		    die "Cannot read $Filename for tailing: $!\n";
		}

		# print a short status message and flush any errors so far to the file.
		logMessage("NMIS_LogWatch forked successfully and tailing $Filename debug=$debug\n");

		# Wrap the actual logwatch code in a loop here:
		while (defined($line=$File->read) and $time_to_die == 0 ) { # will block max 60 secs

		    chomp $line;

			if ($line =~ /Tunnel\d+ from FULL to DOWN/) {
				log_event("down", "ospf");
				next;
		    }
		    if ($line =~ /Tunnel\d+ from LOADING to FULL/) {
				log_event("up", "ospf");
				next;
		    }
		    if ($line =~ /\(Tunnel\d+\) is down/) {
				log_event("down", "eigrp");
				next;
		    }
		    if ($line =~ /\(Tunnel\d+\) is up/) {
				log_event("up", "eigrp");
				next;
		    }
		    if ($line =~ /ENVM-\d+-ENVWARN/) {
				sendTrap();
				next;
		    }
		    
		} # end of while loop
	} # end of child
    elsif ($! == EAGAIN) {
        # EAGAIN is the supposedly recoverable fork error
        sleep 5;
        redo FORK;
    }
    else {
        # weird fork error
        die "Can't fork: $!\n";
    }
}

sub signal_handler {
	my $signame = shift;
	our $time_to_die++;
	my $time = scalar localtime();
	die "$time NMIS_LogWatch - I have died - Somebody sent me a SIG$signame!\n";
}

sub log_event {
	# sample log
	# Jun  9 06:55:49 router.domain.com--f0-0 21773: Jun  9 07:54:38.875 GMT: %OSPF-5-ADJCHG: Process 2000, Nbr 127.0.0.1 on Tunnel15 from FULL to DOWN, Neighbor Down: Dead timer expired
	# Jun  9 07:38:21 router.domain.com--f0-0 45121: Jun  9 08:37:11.244 GMT: %OSPF-5-ADJCHG: Process 2000, Nbr 127.0.0.1 on Tunnel15 from LOADING to FULL, Loading Done
	# pick out the node and the Tunnel##

	my @logline;
	my $node;
	my $interface;

	my $state = shift;
	my $type = shift;

	if ( $line =~ /%/ and $line =~ /:/ ) {		# cisco logline
		$line =~ s/  / /g;	
		@logline = split / / ,$line;

		$node = $logline[3];
		($node) = split /--/ ,$node;	# first split is the one we want
	
		if ( $line =~ /(Tunnel\d+)/ ) {	# get the interface ( Tunnel) number
			$interface = $1;
		}
		else {
			($debug) && logMessage("NMIS_LogWatch No Tunnel## found: $line\n");
			return(0);
		}
	}
	else {
			($debug) && logMessage("NMIS_LogWatch Not a valid logline: $line\n");
			return(0);
	}

 
	# load system file
	if ( $node !~ /\./ ) { # is it a fqdn ?
		$node = $node . "." . $NMIS::config{domain_name};
	}

	if ( -f "$NMIS::config{'<nmis_var>'}/$node.dat" ) {
		loadSystemFile($node);
	}
	else {	# not a known node
		($debug) && logMessage("NMIS_LogWatch $node.dat not found\n");
		return(0);
	}

	# load interface file so we can report the interface description as well to the event log - handy for sysadmin.

	# See if an interface definition file exists and get the description, if this function enabled
	if ( -r "$NMIS::config{'<nmis_var>'}/$node-interface.dat" and $NMIS::config{send_description} eq "true"	) {
		($debug) && logMessage("NMIS_LogWatch getting $interface description from $NMIS::config{'<nmis_var>'}/$node-interface.dat\n");
		my %interfaceTable = &loadCSV("$NMIS::config{'<nmis_var>'}/$node-interface.dat","ifDescr","\t");

		if ( defined $interfaceTable{lc($interface)}{Description} and $interfaceTable{lc($interface)}{Description} ne "null" ) {
			$interface .= " description=$interfaceTable{lc($interface)}{Description}";
		}
	}

	# Tunnel down
	if ( $state eq "down" ) {
		notify(node => $node, role => $NMIS::systemTable{roleType}, type => $NMIS::systemTable{nodeType}, event => "Interface Down", details => "interface=$interface");
		($debug) && logMessage("NMIS_LogWatch node=$node, type=$type, event=interface down, details=$interface\n");
	}


	# Tunnel Up
	if ( $state eq "up" ) {
		checkEvent(node => $node, role => $NMIS::systemTable{roleType}, type => $NMIS::systemTable{nodeType}, event => "Interface Down", details => "interface=$interface");
		($debug) && logMessage("NMIS_LogWatch node=$node, type=$type, event=interface up, details=$interface\n");
	}
}

# if we get an ENV warning - log as a trap - will need to be deleted once ack - use event editor to do that - new to view.pl Dec2004

sub sendTrap {

	my @logline;
	my $node;

	if ( $line =~ /%/ and $line =~ /:/ ) {		# cisco logline
		$line =~ s/  / /g;	
		@logline = split / / ,$line;

		$node = $logline[3];
		($node) = split /--/ ,$node;	# first split is the one we want
	}
	else {
			($debug) && logMessage("NMIS_LogWatch Not a valid logline: $line\n");
			return(0);
	}

 
	# load system file
	if ( $node !~ /\./ ) { # is it a fqdn ?
		$node = $node . "." . $NMIS::config{domain_name};
	}

	if ( -f "$NMIS::config{'<nmis_var>'}/$node.dat" ) {
		loadSystemFile($node);
	}
	else {	# not a known node
		($debug) && logMessage("NMIS_LogWatch $node.dat not found\n");
		return(0);
	}

	# send the whole logline to the event system, so HelpDesk can see what is going on..
	notify(node => $node, role => $NMIS::systemTable{roleType}, type => $NMIS::systemTable{nodeType}, event => "TRAP", details => "$line");
	($debug) && logMessage("NMIS_LogWatch node=$node, event=TRAP, details=$line\n");
}
