#!/usr/bin/perl
#
#    nbarpdd.pl - NMIS Perl Program - Network Mangement Information System
#    Copyright (C) 2000 Sinclair InterNetworking Services Pty Ltd
#    <nmis@sins.com.au> http://www.sins.com.au/nmis
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#*****************************************************************************
#
# written by Cologne 2005
#
# run this from the <nmis>/bin directory, normaly it will be start by nmis.pl
# turn on debug if you want detailed logging to the <nmis>/logs/nmis.log file
# will fork and leave a status message so you know it is running.
#
# Auto configure to the <nmis-base>/lib and <nmis-base>/conf/nmis.conf
#

require 5.008_000;

use FindBin;
use lib "$FindBin::Bin/../lib";
use strict;
use BER;
use SNMP_Session '1.08';
use SNMP_Simple;
use SNMPv2c_Simple;
use csv;
use func;
use NMIS;
use POSIX;
use RRDs 1.000.490;
use rrdfunc;

use Data::Dumper; 
$Data::Dumper::Indent = 1;

# this imports the LOCK_ *constants (eg. LOCK_UN, LOCK_EX)
use Fcntl qw(:DEFAULT :flock);

my $pid;
my $line;
my $time_to_die = 0;

# Variables for command line munging
my %nvp = getArguements(@ARGV);

# See if customised config file required.
my $conf;
if ( $nvp{file} ne "" ) { $conf = $nvp{file}; }
else { $conf = "nmis.conf"; }
my $configfile = "$FindBin::Bin/../conf/$conf";
if ( -r $configfile ) { loadConfiguration($configfile); }
# the following should be conformant to Linux FHS
elsif ( -e "/etc/nmis/$conf") { loadConfiguration ("/etc/nmis/$conf");}
else { die "Can't access neither configuration file $configfile, nor /etc/nmis/$conf \n"; }

# Set debugging level.
my $debug = setDebug($nvp{debug});
my $vardir = $nvp{vardir};

unless ( $vardir ) { $vardir = $NMIS::config{'<nmis_var>'} }

# load mib
foreach ( split /,/ , $NMIS::config{full_mib} ) {
	if ( ! -r "$NMIS::config{mib_root}/$_" ) { 
		 warn returnTime." NBARPDD.pl, mib file $NMIS::config{mib_root}/$_ not found.\n";
	}
	else {
		SNMP_MIB::loadoids_file( $NMIS::config{mib_root}, $_ );
		if ($debug) { print "\t Loaded mib $NMIS::config{mib_root}/$_\n"; }
		if ( $SNMP_Simple::errmsg ) {
			warn returnTime." NBARPDD.pl, SNMP error. errmsg=$SNMP_Simple::errmsg\n";
			$SNMP_Simple::errmsg = "";
		}
	}
}

# Initialize globals

my $restartcmd = "$0 '$vardir'";
my $collect_time = 60; # collect interval time in seconds
my %NBARPDcfg; # global, configuration values
my %NBARPDdata; # global, data values
my %PDprev; # global, previous collected values
my %PDinfo; # global, previous collected values
my %PDtable;
my $BLKindx = 100; # number of indexes read by snmp bulk

# RRD sizing
my $RRD_poll = $collect_time; # default 60 sec.
my $RRD_hbeat = $RRD_poll * 3;
my $RRA_step_1 = 1;
my $RRA_rows_1 = 2880; # 2 days
my $RRA_step_2 = 5;
my $RRA_rows_2 = 288; # 1 days multiplied by value history

$SNMP_Simple::suppress_warnings = 1;
$SNMP_Session::suppress_warnings = 1;

# See if another instance of ourself is running, if so kill the process
my $pidfile = $vardir."/nbarpdd.pid";

if (-f $pidfile) {
  open(F, "<$pidfile");
  $pid = <F>;
  close(F);
  chomp $pid;
  if ($pid != $$) {
    logMessage("NBARPDD pidfile exists killing the pidfile process $pid");
    kill 9, $pid;
    unlink($pidfile);
    logMessage("NBARPDD pidfile $pidfile deleted");
  }
}

if ( $nvp{type} eq "update") {
	logMessage("NBARPDD: update request from nbarpd.pl received") if $debug;
	runPD(0);
	exit 0;
}

# As we are paranoid and running as root, chroot to a safe directory:
# note: if we go anywhere else that will break the file paths.
chroot("/") or die "Couldn't chroot to /: $!";


FORK: {
    if ($pid = fork) {
        # parent here
        # child process pid is available in $pid
		exit;	
    }
    elsif (defined $pid) { # $pid is zero here if defined
        # child here
        # parent process pid is available with getppid

		# dissociate from the controlling terminal
		POSIX::setsid() or die "Can't start new session: $!"; 
		# set our named so that rc scripts can figure out who we are.
		$0="NMIS nbarpdd debug=$debug";
		
		# Announce our presence via a PID file

		open(PID, ">$pidfile") || exit;
		print PID $$; close(PID);
		logMessage("NBARPDD: pidfile $pidfile created");

		# Perform a sanity check. If the current PID file is not the same as
		# our PID then we have become detached somehow, so just exit

		open(PID, "<$pidfile") || exit;
		$pid = <PID>; close(PID);
		chomp $pid;
		exit unless $pid == $$;

		# Record our (re)starting in the event log

		logMessage("NBARPDD: start: pidfile=$pidfile pid=$pid");

		# code body here.

		# setup a log file and trap any error messages as well.
		umask 0;
		open STDOUT, ">>$NMIS::config{log_root}/nmis.log" or die "Can't write to $NMIS::config{log_root}/nmis.log: $!";
		open STDERR, ">>$NMIS::config{log_root}/nmis.log" or die "Can't write to $NMIS::config{log_root}/nmis.log: $!";

		# setup a trap for fatal signals, setting a flag to indicate we need to gracefully exit.
		$SIG{INT} = $SIG{TERM} = $SIG{HUP} = \&signal_handler;
		$SIG{'CHLD'} = 'IGNORE';		# autoreap zombies

		# print a short status message and flush any errors so far to the file.
		logMessage("NBARPDD: forked successfully debug=$debug\n");

		# 
		my $time1;
		my $time2;
		while ($time_to_die == 0 ) { 
			$time1 = time();
			open(PID, '+<',"$pidfile"); print PID $$; close(PID); # signal client that we are running again
			runPD(1);
			$time2 = time(); 
			if ( ($time2 - $time1) > $collect_time ) { logMessage("NBARPDD: runPD, runtime of collecting exceed collect interval time\n"); }
			sleep( $collect_time - ($time2 % $collect_time) );
		} # end of while loop
	} # end of child
    elsif ($! == EAGAIN) {
        # EAGAIN is the supposedly recoverable fork error
        sleep 5;
        redo FORK;
    }
    else {
        # weird fork error
        die "NBARPDD: Can't fork: $!\n";
    }
}

sub signal_handler {
	my $signame = shift;
	our $time_to_die++;
	my $time = scalar localtime();
	die "NBARPDD: - I have died - Somebody sent me a SIG$signame!\n";
}

#========

sub runPD {
	my $collect = shift;

	my $statusPDenable;
	my $runtime;

	if ($collect) {
		$runtime = time(); # start of collecting
		%NBARPDdata = (); # empty global hash
	}

	logMessage("NBARPDD: runPD, start collect with $collect\n") if $debug;

	# read config
	%NBARPDcfg = readVartoHash("nbarpdcfg"); # global hash
	loadNodeDetails(); # %NMIS::nodeTable{} global

	my @keys = keys %NBARPDcfg;
	foreach my $ni (@keys) {
		if (exists $NBARPDcfg{$ni}{node} and $NBARPDcfg{$ni}{node} ne ""){
			logMessage("NBARPDD: runPD, key $ni, function ($NBARPDcfg{$ni}{func}), status ($NBARPDcfg{$ni}{status})\n") if $debug;
			if ($NBARPDcfg{$ni}{func} =~ /start|remove|stop/) {
				my $sessionRO;
				# open RO session
				$SNMP_Simple::errmsg = '';
				($sessionRO) = SNMPv2c_Simple->open($NBARPDcfg{$ni}{node}, $NMIS::nodeTable{$NBARPDcfg{$ni}{node}}{community}, $NMIS::nodeTable{$NBARPDcfg{$ni}{node}}{snmpport});
				if ( not defined($sessionRO)) { 
					$NBARPDcfg{$ni}{status} = "error" ;
					$NBARPDcfg{$ni}{message} = "SNMP session is not Defined Goodly to $NBARPDcfg{$ni}{node}";
					logMessage("NBARPDD: runPD, $NBARPDcfg{$ni}{message}");
				} else {
					# request nbarpd status of interface
					($statusPDenable) = $sessionRO->snmpget("cnpdStatusPdEnable.$NBARPDcfg{$ni}{intf}");
					if ( $SNMP_Simple::errmsg =~ /No answer|Unknown/ and $NBARPDcfg{$ni}{status} =~ /start/) {
						$NBARPDcfg{$ni}{status} = "error" ;
						$NBARPDcfg{$ni}{message} = $SNMP_Simple::errmsg ;
						logMessage("NBARPDD: runPD, $SNMP_Simple::errmsg") if $debug;
					} else {
						$sessionRO->close(); $sessionRO = undef;
						if ($NBARPDcfg{$ni}{func} eq "start") {
							&runPDstart($ni,$statusPDenable);

						} elsif ($NBARPDcfg{$ni}{func} eq "stop" or $NBARPDcfg{$ni}{func} eq "remove") {
							&runPDstop($ni,$statusPDenable);

						}
					}
				}
				$NBARPDcfg{$ni}{func} = ""; # done
				runPDmodify($ni); # store modified config
				if (defined $sessionRO) { $sessionRO->close(); }
			}
			if ($NBARPDcfg{$ni}{status} eq "running" and $collect) {
				my $sessionRO;
				# open session
				$SNMP_Simple::errmsg = '';
				$sessionRO = SNMPv2c_Session->open ($NBARPDcfg{$ni}{node}, $NMIS::nodeTable{$NBARPDcfg{$ni}{node}}{community}, $NMIS::nodeTable{$NBARPDcfg{$ni}{node}}{snmpport});
				if ( not defined $sessionRO) { 
					$NBARPDcfg{$ni}{status} = "error" ;
					$NBARPDcfg{$ni}{message} = "SNMP session is not Defined Goodly to $NBARPDcfg{$ni}{node}";
					logMessage("NBARPDD: runPD, SNMP session is not Defined Goodly to $NBARPDcfg{$ni}{node}");
					runPDmodify($ni); # store config
				} else {
					runPDrunning($sessionRO,$ni);
				}

				if (defined $sessionRO) { $sessionRO->close(); }
			}
		}
	}

	if ($collect) {
		$NBARPDdata{runtime} = time() - $runtime;
		writeHashtoVar("nbarpddata",\%NBARPDdata); # store data for web page pie and bar
	}

}

sub runPDstart {
	my $ni = shift;
	my $statusPDenable = shift;
	my $sessionRW;

	# check if interface already running nbarpd
	if ($statusPDenable eq "1") {
		$NBARPDcfg{$ni}{status} = "running" ;
		$NBARPDcfg{$ni}{message} = "";
		$NBARPDcfg{$ni}{enabled} = "true"; # already enabled
		$NBARPDcfg{$ni}{starttime} = "started at ".returnDateStamp().", sample rate $collect_time sec. (nbarpd already enabled on interface)";
		logMessage("NBARPDD: runPDstart, $NBARPDcfg{$ni}{node}::$NBARPDcfg{$ni}{descr} already running\n") if $debug;
	} else {
		if ($NBARPDcfg{$NBARPDcfg{$ni}{node}}{community} eq "") {
			$NBARPDcfg{$ni}{status} = "error" ;
			$NBARPDcfg{$ni}{message} = "cannot enable node::interface, no community RW defined";
			logMessage("NBARPDD: runPDstart, SNMP session is not Defined Goodly to $NBARPDcfg{$ni}{node}");
		} else {
			# open RW session
			($sessionRW) = SNMPv2c_Simple->open($NBARPDcfg{$ni}{node}, $NBARPDcfg{$NBARPDcfg{$ni}{node}}{community}, $NMIS::nodeTable{$NBARPDcfg{$ni}{node}}{snmpport});
			if ( not defined($sessionRW)) { 
				$NBARPDcfg{$ni}{status} = "error" ;
				$NBARPDcfg{$ni}{message} = "SNMP session is not Defined Goodly to $NBARPDcfg{$ni}{node}";
				logMessage("NBARPDD: runPDstart, SNMP session is not Defined Goodly to $NBARPDcfg{$ni}{node}");
			} else {
				# oke, start PD on interface
				$sessionRW->snmpset("cnpdStatusPdEnable.$NBARPDcfg{$ni}{intf}","int","1");
				if ( $SNMP_Simple::errmsg =~ /No answer|Unknown/ ) {
					$NBARPDcfg{$ni}{status} = "error" ;
					$NBARPDcfg{$ni}{message} = $SNMP_Simple::errmsg ;
					logMessage("NBARPDD: runPDstart, $SNMP_Simple::errmsg") if $debug;
				} else {
					$NBARPDcfg{$ni}{status} = "running" ;
					$NBARPDcfg{$ni}{message} = "";
					$NBARPDcfg{$ni}{starttime} = "started at ".returnDateStamp().", sample rate $collect_time sec.";
					logMessage("NBARPDD: runPDstart, $NBARPDcfg{$ni}{node}::$NBARPDcfg{$ni}{descr} started\n") if $debug;
				}
			}
		}
	}
	if (defined $sessionRW) { $sessionRW->close(); }
}

sub runPDstop {
	my $ni = shift;
	my $statusPDenable = shift;
	my $sessionRW;

	# if nbarpd was already enabled on this interface dont stop it
	if ( $NBARPDcfg{$ni}{enabled} eq "true") {
		$NBARPDcfg{$ni}{status} = "remove";
	} elsif ($statusPDenable eq "2") { # check if interface was already stopped
		# remove configuration
		logMessage("NBARPDD: runPDstop, $NBARPDcfg{$ni}{node}::$NBARPDcfg{$ni}{descr} already stopped\n") if $debug;
		$NBARPDcfg{$ni}{status} = "remove";
	} elsif ($NBARPDcfg{$NBARPDcfg{$ni}{node}}{community} eq "") {
		$NBARPDcfg{$ni}{status} = "error" ;
		$NBARPDcfg{$ni}{message} = "cannot disable nbarpd on this interface, no community RW defined";
		logMessage("NBARPDD: runPDstart,cannot disable nbarpd on $NBARPDcfg{$ni}{node}::$NBARPDcfg{$ni}{descr}, no community RW defined") if $debug;
	} else {
		# open session
		$sessionRW = SNMPv2c_Simple->open($NBARPDcfg{$ni}{node}, $NBARPDcfg{$NBARPDcfg{$ni}{node}}{community}, $NMIS::nodeTable{$NBARPDcfg{$ni}{node}}{snmpport});
		if ( not defined $sessionRW ) { 
			$NBARPDcfg{$ni}{status} = "error" ;
			$NBARPDcfg{$ni}{message} = "SNMP session is not Defined Goodly to $NBARPDcfg{$ni}{node}";
			logMessage("NBARPDD: runPDstop, SNMP session is not Defined Goodly to $NBARPDcfg{$ni}{node}") if $debug;
		} else {
			# stop PD on interface
			$sessionRW->snmpset("cnpdStatusPdEnable.$NBARPDcfg{$ni}{intf}","int","2");
			if ( $SNMP_Simple::errmsg =~ /No answer|Unknown/ and $NBARPDcfg{$ni}{func} eq "stop") {
				$NBARPDcfg{$ni}{status} = "error" ;
				$NBARPDcfg{$ni}{message} = $SNMP_Simple::errmsg;
			} else {
				# remove configuration
				logMessage("NBARPDD: runPDstop, $NBARPDcfg{$ni}{node}::$NBARPDcfg{$ni}{descr} stopped\n") if $debug;
				$NBARPDcfg{$ni}{status} = "remove";
			}
		}
	}
	delete $PDprev{$ni}; # remove history
	delete $PDinfo{$ni}; # remove history
	my $database = "$NMIS::config{'<nmis_var>'}/nbarpd-$ni.rrd";
	unlink $database; # cleanup database
	if (defined $sessionRW) { $sessionRW->close(); }
}

sub runPDmodify {
	my $ni = shift;

	# get NBARPDcfg from /var
	my %PDcfg = readVartoHash("nbarpdcfg");

	if ($NBARPDcfg{$ni}{status} eq "remove") {
		delete $PDcfg{$ni};
		delete $NBARPDcfg{$ni};
	} else {
		$PDcfg{$ni}{func} =$NBARPDcfg{$ni}{func};
		$PDcfg{$ni}{message} =$NBARPDcfg{$ni}{message};
		$PDcfg{$ni}{enabled} =$NBARPDcfg{$ni}{enabled};
		$PDcfg{$ni}{status} =$NBARPDcfg{$ni}{status};
		$PDcfg{$ni}{starttime} =$NBARPDcfg{$ni}{starttime};
	}
	writeHashtoVar("nbarpdcfg",\%PDcfg);
	logMessage("NBARPDD: runPDmodify, write modified config of node-intf $ni to /var\n") if $debug;
}

sub runPDrunning {
	my $session = shift; # sessionRO
	my $ni = shift; # node name and interface number

	my %PDvalues;
	my %PDvaluesName;
	my %PDvaluesInBytes;
	my %PDvaluesOutBytes;
	my $PDvalue_in;
	my $PDvalue_out;
	my @options;
	my $optname;
	my $optrate;
	my $name;
	my $Error;

	# get values from node, returned index and value in hash
	if (not (%PDvaluesInBytes = runPDvalues($session,$ni,"cnpdAllStatsHCInBytes"))) { return;}
	if (not (%PDvaluesOutBytes = runPDvalues($session,$ni,"cnpdAllStatsHCOutBytes"))) { return;}

	my $time = time();
	if ($PDinfo{$ni}{time} eq ""){
		# first collect
		if (not (%PDvaluesName = runPDvalues($session,$ni,"cnpdAllStatsProtocolName"))) { return;}
		foreach my $i (keys %PDvaluesName) {
			$PDprev{$ni}{$i}{name} = $PDvaluesName{$i};
			$PDprev{$ni}{$i}{in} = $PDvaluesInBytes{$i}; 
			$PDprev{$ni}{$i}{out} = $PDvaluesOutBytes{$i};
			$PDinfo{$ni}{rows} = $i if $i > $PDinfo{$ni}{rows}; # the highest one
		}
		$PDinfo{$ni}{time} = $time;
		writeHashtoVar("nbarpd-prev-$ni",\%PDprev) if $debug;
	} else {
		# check if the number of protocol names has been changed (by IOS config)
		my $rowsold = $PDinfo{$ni}{rows} - 1;
		my $rowsnew = scalar(keys %PDvaluesInBytes);
		if ( $rowsnew != $rowsold) {
			# its different, perhaps one or more new protocol names are configured by IOS
			%PDvaluesName = runPDvalues($session,$ni,"cnpdAllStatsProtocolName"); # get protocol names
			# find the protocol names in the existing database (format ds[name].* )
			my $database = "$NMIS::config{'<nmis_var>'}/nbarpd-$ni.rrd";
			my $info = RRDs::info($database);
			my @names;
			foreach my $key (keys %$info) {
				if ( $key =~ /^ds\[([a-zA-Z0-9_-]{1,16})_in\].+/) { 
					my $nm = $1;
					if ($nm ne "" and ! grep /$nm/, @names ) { push @names,"$nm";} 
				}
			}
			# which protocol name is missing in database
			foreach my $i (keys %PDvaluesName) {
				if ( ! grep /$PDvaluesName{$i}/ , @names) {
					# found missing name
					my $name = $PDvaluesName{$i};
					logMessage("NBARPDD: runPDrunning, new protocol name $name found, add to database");
					addDStoRRD($database,"DS:${name}_in:GAUGE:$RRD_hbeat:U:U");
					addDStoRRD($database,"DS:${name}_out:GAUGE:$RRD_hbeat:U:U");
				}
			}
			# rebuild PDprev table
			delete $PDprev{$ni};
			delete $PDinfo{$ni};
			foreach my $i (keys %PDvaluesName) {
				$PDprev{$ni}{$i}{name} = $PDvaluesName{$i};
				$PDprev{$ni}{$i}{in} = eval $PDvaluesInBytes{$i};
				$PDprev{$ni}{$i}{out} = eval $PDvaluesOutBytes{$i};
				$PDinfo{$ni}{rows} = $i if $i > $PDinfo{$ni}{rows}; # the highest one
			}
			$PDinfo{$ni}{time} = $time;
			$NBARPDcfg{$ni}{message} = "number of protocol names has been changed, wait for rebuild";
			runPDmodify($ni);
		} elsif ( $PDprev{$ni}{time} >= $time ) {
			foreach my $i (keys %PDvaluesInBytes) {
				$PDprev{$ni}{$i}{in} = $PDvaluesInBytes{$i}; # store for next run
				$PDprev{$ni}{$i}{out} = $PDvaluesOutBytes{$i};
			}
			$NBARPDcfg{$ni}{message} = "clock changed, previous sample time is in the future";
			logMessage("NBARPDD: runPDrunning, clock changed, previous sample time is in the future") if $debug;
			runPDmodify($ni);
		} else {
			# calculate and store values 
			foreach my $i (keys %PDvaluesInBytes) {
				my $name = $PDprev{$ni}{$i}{name};
				# calculate bitrate and remove Math::BigInt object
				$PDvalue_in = eval int((($PDvaluesInBytes{$i} - $PDprev{$ni}{$i}{in}) * 8) / ($time - $PDprev{$ni}{time})); # bitrate
				$PDvalue_out = eval int((($PDvaluesOutBytes{$i} - $PDprev{$ni}{$i}{out}) * 8) / ($time - $PDprev{$ni}{time})); # bitrate
				# build option scalar for rrd
				if ($optname eq "") { 
					$optname = "${name}_in:${name}_out";
					$optrate = "N:$PDvalue_in:$PDvalue_out";
				} else { 
					$optname .= ":${name}_in:${name}_out"; 
					$optrate .= ":$PDvalue_in:$PDvalue_out";
				}
				$PDprev{$ni}{$i}{in} = $PDvaluesInBytes{$i}; # store for next run
				$PDprev{$ni}{$i}{out} = $PDvaluesOutBytes{$i};

				# only != zero for pie and bar graphs
				if ($PDvalue_in != 0 or $PDvalue_out != 0) {
					$PDvalues{$ni}{$i}{name} = $name ;
					$PDvalues{$ni}{$i}{inrate} = $PDvalue_in ; 
					$PDvalues{$ni}{$i}{outrate} = $PDvalue_out ;
				}
			}
			$PDprev{$ni}{time} = $time; # remember collect time to calculate bitrate
			$PDvalues{$ni}{'000'}{lastsample} = $time; # for web page

			# combine data in one hash
			%NBARPDdata = (%NBARPDdata,%PDvalues);

			# store data for every node::interface in rrd
			my $database = "$NMIS::config{'<nmis_var>'}/nbarpd-$ni.rrd";
			# Does the database exist already?
			if ( not (-f $database and -r $database and -w $database) ) { 
				%PDvaluesName = runPDvalues($session,$ni,"cnpdAllStatsProtocolName"); # get protocol names
				if ( not runRRDcreate($ni,$database,\%PDvaluesName)) { return;}
			}

			@options = ( "-t", $optname, $optrate);

			# update RRD
			RRDs::update($database,@options);
			if ($Error = RRDs::error) {
				$NBARPDcfg{$ni}{status} = "error" ;
				$NBARPDcfg{$ni}{message} = "update rrd database, $Error";
				logMessage("NBARPDD: runPDrunning, update database $database, $Error\n") if $debug;
				runPDmodify($ni);
			} else {
				logMessage("NBARPDD: runPDrunning, updated database $database\n") if $debug;
				if ($NBARPDcfg{$ni}{message} ne "") { $NBARPDcfg{$ni}{message} = ""; runPDmodify($ni); }
			}
		}
	}

#==
	# read values from node by snmp bulk
	sub runPDvalues {
		my $session = shift; # open session descr.
		my $ni = shift;	# $node-$interface
		my $name = shift; # oid

		%PDtable = (); # empty
		$SNMP_Session::errmsg = "";

	    my @name_oid = split( /\./, SNMP_MIB::name2oid($name).".".$NBARPDcfg{$ni}{intf});

		$session->map_table_start_end([\@name_oid],\&out_values,1,$BLKindx,40);

		if ($SNMP_Session::errmsg =~ /no response/i ) {
			# perhaps temporary unreachable
			if ($NBARPDcfg{$ni}{message} eq "") {
				$NBARPDcfg{$ni}{message} = "node $NBARPDcfg{$ni}{node} unreachable";
				runPDmodify($ni); # update config file
			}
			return;
		}
		if ( $SNMP_Session::errmsg ne "" ) {
			$NBARPDcfg{$ni}{status} = "error" ;
			$NBARPDcfg{$ni}{message} = $SNMP_Session::errmsg;
			logMessage("NBARPDD: runPDvalues, $SNMP_Session::errmsg\n") if $debug;
			runPDmodify($ni);
			return;
		}

		if (not %PDtable) {
			if (not ($NBARPDcfg{$ni}{status} eq "true")) { runPDstart( $ni, 2); } # try this first
			if ($NBARPDcfg{$ni}{message} eq "" and $NBARPDcfg{$ni}{status} eq "true") {
				$NBARPDcfg{$ni}{message} = "no info from node $NBARPDcfg{$ni}{node} received, perhaps nbarpd disabled on interface";
				runPDmodify($ni);
			}
			return;
		}

		if ($NBARPDcfg{$ni}{message} ne "") {
			$NBARPDcfg{$ni}{message} = ""; # clear msg
			runPDmodify($ni);
		}

		return %PDtable;
	#==
		sub out_values {
			my ($index, $value) = @_;
			grep (defined $_ && ($_=pretty_print $_),($value));
			$PDtable{$index} = $value; # store in global hash 
		}
	}

#==

	sub runRRDcreate {
		my $ni = shift;		# $node-$interface
		my $database = shift; # name of database
		my $PDname = shift; # address of hash
		my $name;

		my $time  = time() - 1;
		my @options;

		@options = ( "-b", $time, "-s", $RRD_poll );

		foreach my $i (keys %{$PDname}) {
			$name = $$PDname{$i};
			push @options, "DS:${name}_in:GAUGE:$RRD_hbeat:U:U";
			push @options, "DS:${name}_out:GAUGE:$RRD_hbeat:U:U";
		}

		my $rows_2 = ($RRA_rows_2 * 7) * $NBARPDcfg{$ni}{history}; # calc. db size
		push @options, "RRA:AVERAGE:0.5:$RRA_step_1:$RRA_rows_1";
		push @options, "RRA:AVERAGE:0.5:$RRA_step_2:$rows_2";

		RRDs::create("$database",@options);
		my $ERROR = RRDs::error;
		if ($ERROR) {
			$NBARPDcfg{$ni}{status} = "error" ;
			$NBARPDcfg{$ni}{message} = "create rrd database, $ERROR";
			logMessage("NBARPDD: runRRDcreate, unable to create $database: $ERROR");
			runPDmodify($ni); # update config file
			return 0;
		} else {
			logMessage("NBARPDD: runRRDcreate, created database $database\n") if $debug;
		}
		# set file owner and permission, default: nmis, 0775.
		setFileProt($database);
		return 1;
	}

}


